"""
Represents a single assembly instruction with its operands, location and optional branch
target.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .address import Address


class Coverage(Enum):
    """Enumeration for coverage records."""

    MISS = "missed"
    """Indicates instruction is not executed"""
    LINE_TAKEN = "taken"
    """Indicates instruction is executed"""
    JUMP_PASS = "skipped"
    """Indicates, if branch is just passed without jump"""
    JUMP_TAKEN = "jumped"
    """Indicates, if branch is just jumped without passed"""
    JUMP_BOTH = "both taken"
    """Indicates, if branch is just jumped without passed"""


class Instruction:
    """Instruction Class."""

    lineno: int | None = None
    """Line number in the file"""

    address: Address | None = None
    """Computed address of the instruction"""

    opcode: str | None = None
    """Disassembled opcode"""

    ops: list[str] = []
    """Disassembled operands of the instruction"""

    target: Address | None = None
    """Optional target of the instruction (branch)"""

    returns: set[int] = set()
    """Stores addresses of sink instructions like return"""

    coverage: Coverage = Coverage.MISS
    """If line is executed on test"""

    branch_taken: bool | None = None

    def __init__(
        self,
        lineno: int,
        address: Address,
        opcode: str,
        ops: list[str],
        target: Address | None,
    ) -> None:
        self.lineno = lineno
        self.address = address
        self.opcode = opcode
        self.ops = ops
        self.target = target
        self.returns = set()
        self.coverage = Coverage.MISS

    def update_coverage(self, addresses: set[int], is_branch: bool = False) -> None:
        """
        Update the coverage information of the instruction.

        :param addresses: Set of addresses executed after this instruction,
            used to determine whether a branch was taken or not.
        :param is_branch: Flag whether the instruction should be
            interpreted as a branch
        """
        if not is_branch:
            # exception for ret (target is None) or call (target.abs is None)
            self.coverage = Coverage.LINE_TAKEN
            self.returns = addresses
        elif len(addresses) == 2 and self.target and self.target.abs_addr in addresses:
            self.coverage = Coverage.JUMP_BOTH
        elif len(addresses) == 1 and self.target and self.target.abs_addr in addresses:
            self.coverage = Coverage.JUMP_TAKEN
        elif len(addresses) == 1:
            self.coverage = Coverage.JUMP_PASS
        else:
            msg = f"Invalid Coverage Information at {self.address}: {addresses}"
            raise AssertionError(msg)

    def __repr__(self) -> str:
        """Convert an Instruction into a representative string."""
        result = f"{self.address}: {self.opcode}"
        if self.ops:
            result += f" {self.ops}"
        if self.target is not None:
            result += " -> " + str(self.target)
        return f"Instruction({result})"
